{
  "type": "record",
  "name": "DataHubSecretKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a DataHub Secret",
  "fields": [
    {
      "name": "id",
      "type": "string",
      "doc": "A unique id for the Secret"
    }
  ],
  "Aspect": {
    "name": "dataHubSecretKey",
    "keyForEntity": "dataHubSecret",
    "entityCategory": "internal",
    "entityAspects": [
      "dataHubSecretValue"
    ]
  }
}